# -*- coding: utf-8 -*-

MAJOR_VERSION = 0
MINOR_VERSION = 2
PATCH_VERSION = 0
__short_version__ = "{}.{}".format(MAJOR_VERSION, MINOR_VERSION)
__version__ = "{}.{}".format(__short_version__, PATCH_VERSION)
